@@ -7,7 +7,7 @@ use Carp ();
 use IO::Handle ();
 use Exporter 'import';
 
-our $VERSION = '0.007';
+our $VERSION = '0.008';
 
 use constant DEFAULT_REQUEST_BODY_LIMIT => 16777216;
 use constant DEFAULT_REQUEST_BODY_BUFFER => 131072;
@@ -427,6 +427,8 @@ sub _body_multipart {
 sub _parse_multipart {
   my ($input, $length, $boundary, $buffer_size) = @_;
   my $buffer = "\r\n";
+  my $next_boundary = "\r\n--$boundary\r\n";
+  my $end_boundary = "\r\n--$boundary--";
   my (%state, @parts);
   READER: while ($length > 0) {
     if (ref $input eq 'SCALAR') {
@@ -439,27 +441,76 @@ sub _parse_multipart {
       $length -= $read;
     }
 
-    if (!$state{started} and (my $pos = index $buffer, "\r\n--$boundary\r\n") >= 0) {
-      substr $buffer, 0, $pos + length($boundary) + 6, '';
-      $state{started} = 1;
-      push @parts, $state{part} = {headers => {}, name => undef, filename => undef, size => 0};
+    unless ($state{parsing_headers} or $state{parsing_body}) {
+      my $next_pos = index $buffer, $next_boundary;
+      my $end_pos = index $buffer, $end_boundary;
+      if ($next_pos >= 0 and ($end_pos < 0 or $end_pos > $next_pos)) {
+        substr $buffer, 0, $next_pos + length($next_boundary), '';
+        $state{parsing_headers} = 1;
+        push @parts, $state{part} = {headers => {}, name => undef, filename => undef, size => 0};
+      } elsif ($end_pos >= 0) {
+        $state{done} = 1;
+        last; # end of multipart data
+      } else {
+        next; # read more to find start of multipart data
+      }
     }
-    next unless $state{started}; # read more to find start of multipart data
 
     while (length $buffer) {
-      if ($state{parsing_body}) {
+      if ($state{parsing_headers}) {
+        while ((my $pos = index $buffer, "\r\n") >= 0) {
+          if ($pos == 0) { # end of headers
+            $state{parsing_headers} = 0;
+            $state{parsing_body} = 1;
+            $state{parsed_optional_crlf} = 0;
+            last;
+          }
+
+          my $header = substr $buffer, 0, $pos + 2, '';
+          my ($name, $value) = split /\s*:\s*/, $header, 2;
+          return undef unless defined $value;
+          $value =~ s/\s*\z//;
+
+          $state{part}{headers}{lc $name} = $value;
+          if (lc $name eq 'content-disposition') {
+            if (my ($name_quoted, $name_unquoted) = $value =~ m/;\s*name\s*=\s*(?:"((?:\\"|[^";])*)"|([^";]*))/i) {
+              $state{part}{name} = defined $name_quoted ? $name_quoted : $name_unquoted;
+            }
+            if (my ($filename_quoted, $filename_unquoted) = $value =~ m/;\s*filename\s*=\s*(?:"((?:\\"|[^"])*)"|([^";]*))/i) {
+              $state{part}{filename} = defined $filename_quoted ? $filename_quoted : $filename_unquoted;
+            }
+          }
+        }
+        next READER if $state{parsing_headers}; # read more to find end of headers
+      } else {
         my $append = '';
-        my $pos;
-        if (($pos = index $buffer, "\r\n--$boundary\r\n") >= 0) {
-          $append = substr $buffer, 0, $pos, '';
-          substr $buffer, 0, length($boundary) + 6, '';
+        my $next_pos = index $buffer, $next_boundary;
+        my $end_pos = index $buffer, $end_boundary;
+        if ($next_pos >= 0 and ($end_pos < 0 or $end_pos > $next_pos)) {
+          if (!$state{parsed_optional_crlf} and $next_pos >= 2) {
+            substr $buffer, 0, 2, '';
+            $next_pos -= 2;
+            $state{parsed_optional_crlf} = 1;
+          }
+          $append = substr $buffer, 0, $next_pos, '';
+          substr $buffer, 0, length($next_boundary), '';
           $state{parsing_body} = 0;
-        } elsif (($pos = index $buffer, "\r\n--$boundary--") >= 0) {
-          $append = substr $buffer, 0, $pos; # no replacement, we're done here
+          $state{parsing_headers} = 1;
+        } elsif ($end_pos >= 0) {
+          if (!$state{parsed_optional_crlf} and $end_pos >= 2) {
+            substr $buffer, 0, 2, '';
+            $end_pos -= 2;
+            $state{parsed_optional_crlf} = 1;
+          }
+          $append = substr $buffer, 0, $end_pos; # no replacement, we're done here
           $state{parsing_body} = 0;
           $state{done} = 1;
-        } elsif (length($buffer) > length($boundary) + 6) {
-          $append = substr $buffer, 0, length($buffer) - length($boundary) - 6, '';
+        } elsif (length($buffer) > length($next_boundary) + 2) {
+          if (!$state{parsed_optional_crlf}) {
+            substr $buffer, 0, 2, '';
+            $state{parsed_optional_crlf} = 1;
+          }
+          $append = substr $buffer, 0, length($buffer) - length($next_boundary), '';
         }
 
         if (defined $state{part}{filename}) {
@@ -490,30 +541,6 @@ sub _parse_multipart {
 
         # new part started
         push @parts, $state{part} = {headers => {}, name => undef, filename => undef, size => 0};
-      } else { # part headers
-        while ((my $pos = index $buffer, "\r\n") >= 0) {
-          if ($pos == 0) { # end of headers
-            substr $buffer, 0, 2, '';
-            $state{parsing_body} = 1;
-            last;
-          }
-
-          my $header = substr $buffer, 0, $pos + 2, '';
-          my ($name, $value) = split /\s*:\s*/, $header, 2;
-          return undef unless defined $value;
-          $value =~ s/\s*\z//;
-
-          $state{part}{headers}{lc $name} = $value;
-          if (lc $name eq 'content-disposition') {
-            if (my ($name_quoted, $name_unquoted) = $value =~ m/;\s*name\s*=\s*(?:"((?:\\"|[^";])*)"|([^";]*))/i) {
-              $state{part}{name} = defined $name_quoted ? $name_quoted : $name_unquoted;
-            }
-            if (my ($filename_quoted, $filename_unquoted) = $value =~ m/;\s*filename\s*=\s*(?:"((?:\\"|[^"])*)"|([^";]*))/i) {
-              $state{part}{filename} = defined $filename_quoted ? $filename_quoted : $filename_unquoted;
-            }
-          }
-        }
-        next READER unless $state{parsing_body}; # read more to find end of headers
       }
     }
   }
@@ -603,6 +630,14 @@ sub add_response_cookie {
   return $self;
 }
 
+sub response_status_code {
+  my ($self) = @_;
+  if (defined $self->{response_status} and $self->{response_status} =~ m/\A([0-9]+)/) {
+    return 0+$1;
+  }
+  return 200;
+}
+
 sub headers_rendered { $_[0]{headers_rendered} }
 
 my %known_types = (json => 1, html => 1, xml => 1, text => 1, data => 1, redirect => 1);
@@ -643,7 +678,7 @@ sub render {
       $headers_str = "Date: $date_str\r\n$headers_str";
     }
     my $status = $self->{response_status};
-    $status = "302 $HTTP_STATUS{302}" if !defined $status and $type eq 'redirect';
+    $status = $self->{response_status} = "302 $HTTP_STATUS{302}" if !defined $status and $type eq 'redirect';
     if ($self->{nph}) {
       $status = "200 $HTTP_STATUS{200}" unless defined $status;
       my $protocol = $ENV{SERVER_PROTOCOL};
